"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = require("@salesforce/command");
const core_1 = require("@salesforce/core");
const needle = require("needle");
const URL = require("url");
const _ = require("lodash");
const fs = require("fs-extra");
const jsforce = require("jsforce");
var execSync = require("child_process").execSync;
const url_1 = require("url");
// Initialize Messages with the current plugin directory
core_1.Messages.importMessagesDirectory(__dirname);
// Load the specific messages for this file. Messages from @salesforce/command, @salesforce/core,
// or any library that is using the messages framework can also be loaded this way.
// const messages = Messages.loadMessages('fonteva-sfdx', 'org');
class Install extends command_1.SfdxCommand {
    installPackage(pack, targetUserName) {
        try {
            if (!pack) {
                console.error("Package not found");
                return;
            }
            console.log(`Installing ${pack.name}; release: ${pack.versionName}; version: ${pack.version}`);
            if (this.flags.dryrun) {
                return true;
            }
            let urlObj = URL.parse(pack.installLink);
            let packageId = new url_1.URLSearchParams(urlObj.search).get("p0");
            let installCommand = `sfdx force:package:install --package ${packageId} --targetusername ${targetUserName} -r --wait 15 --securitytype AllUsers`;
            console.log(installCommand);
            execSync(installCommand, { stdio: [0, 1, 2] });
            if (!pack.name.match('FDS\\d\\d') && !pack.name.match('KeyStore')) { //FDServiceVersions don't have install scripts
                this.runInstallScript(pack.installClass, targetUserName);
            }
        }
        catch (err) {
            console.log(err);
            return false;
        }
        return true;
    }
    getLatestProductionPackages(packageName, marketplaceResponse) {
        let packageObj = _.filter(marketplaceResponse, { namespace: packageName });
        let latestFullPackage = { versionInt: null };
        packageObj.forEach(function (element) {
            if (latestFullPackage.versionInt == null || latestFullPackage.versionInt < element.versionInt) {
                latestFullPackage = element;
            }
        });
        return latestFullPackage;
    }
    runInstallScript(installScript, targetUserName) {
        console.log('Running Install Script');
        let directory = __dirname + '/../../../../tmp';
        fs.ensureDirSync(directory);
        let fileName = directory + '/installScripts.apex';
        fs.outputFileSync(fileName, installScript + ';');
        let installCommand = `sfdx force:apex:execute --targetusername ${targetUserName} -f "${fileName}"`;
        execSync(installCommand, { stdio: [0, 1, 2] });
    }
    installPackages(self, result) {
        let packageName = self.flags.packagename;
        let url = 'https://fonteva-marketplace-automation.herokuapp.com/api/packages?all=true';
        if (this.flags.isproduction) {
            url = 'https://fonteva-marketplace.herokuapp.com/api/packages?all=true';
        }
        if (this.flags.patch) {
            url = `https://fonteva-marketplace.herokuapp.com/api/package/patch_packages?version=${this.flags.patch}`;
        }
        needle('get', url, function (error, response, body) {
            // local function to loop through a list of dependencies
            const installDependencies = function (packageObj) {
                packageObj.fullDependencies.every(function (element) {
                    let depPack = _.find(body.data, { namespace: element.namespace, version: element.version });
                    if (_.findIndex(result, { fullName: element.namespace, versionNumber: element.version }) === -1) {
                        return self.installPackage(depPack, self.flags.targetusername); // returns false on failure
                    }
                    return true;
                });
            };
            const installPatchDependencies = function (packageObj) {
                packageObj.fullDependencies.every(function (element) {
                    const depPack = _.find(body, { namespace: element.namespace, version: element.version });
                    if (_.findIndex(result, { fullName: element.namespace, versionNumber: element.version }) === -1) {
                        return self.installPackage(depPack, self.flags.targetusername); // returns false on failure
                    }
                    return true;
                });
            };
            if (self.flags.allcoredeps) {
                let packageObj = self.getLatestProductionPackages(packageName, body.data);
                installDependencies(packageObj);
                if (_.findIndex(result, { fullName: packageObj.namespace, versionNumber: packageObj.version }) === -1) {
                    self.installPackage(packageObj, self.flags.targetusername);
                }
            }
            else if (self.args.package) {
                let packageObj = _.find(body.data, { namespace: self.args.package });
                self.installPackage(packageObj, self.flags.targetusername);
            }
            else if (self.flags.patch) {
                let packageObj = _.find(body, { namespace: packageName });
                installPatchDependencies(packageObj);
            }
            else {
                let packageObj = _.find(body.data, { namespace: packageName });
                installDependencies(packageObj);
            }
        });
    }
    async run() {
        if (!this.args.package && !this.flags.packagename) {
            console.error("You must provide either the package argument or the --packagename flag");
            return {};
        }
        const conn = this.org.getConnection();
        const jsConn = new jsforce.Connection({
            serverUrl: conn.instanceUrl,
            sessionId: conn.accessToken
        });
        const self = this;
        jsConn.metadata.read('InstalledPackage', ['Framework', 'KEYSTORE', 'PagesApi', 'OrderApi', 'EventApi', 'ROEApi', 'LTE', 'FDService', 'FDS19R1'], function (err, result) {
            if (result != null && !Array.isArray(result)) {
                result = [result];
            }
            self.installPackages(self, result);
        });
        return {};
        // Return an object to be displayed with --json
        //return { orgId: this.org.getOrgId(), outputString };
    }
}
exports.default = Install;
Install.description = 'Install dependent packages.';
Install.examples = [
    `$ sfdx install:dependencies --targetusername myOrg@example.com --targetdevhubusername devhub@org.com
  Hello world! This is org: MyOrg and I will be around until Tue Mar 20 2018!
  My hub org id is: 00Dxx000000001234
  `
];
Install.args = [{ name: 'package', required: false, description: "(Optional) The Fonteva package to install (e.g. 'Framework')" }];
Install.flagsConfig = {
    packagename: command_1.flags.string({ char: 'p', description: 'Package to install deps for (e.g. "OrderApi" will install PagesApi & Framework)', required: false }),
    isproduction: command_1.flags.boolean({ char: 'd', description: 'If Is Production' }),
    patch: command_1.flags.string({ char: 'P', description: 'If Is Patch - provide the patch args (e.g. --patch=2019.1)' }),
    allcoredeps: command_1.flags.boolean({ char: 'c', description: 'All Core Deps' }),
    dryrun: command_1.flags.boolean({ description: "Dry Run - don't install anything" })
};
// Comment this out if your command does not require an org username
Install.requiresUsername = true;
// Comment this out if your command does not support a hub org username
Install.supportsDevhubUsername = true;
// Set this to true if your command requires a project workspace; 'requiresProject' is false by default
Install.requiresProject = false;
//# sourceMappingURL=dependencies.js.map