"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.doBuild = void 0;
const core_1 = require("@salesforce/core");
const bb = require('bluebird');
const fs = bb.promisifyAll(require('fs-extra'));
const _ = require('lodash');
const child_process = require('child_process');
const archiver = require('archiver');
const convert = require('xml-js');
const path = require('path');
// Initialize Messages with the current plugin directory
core_1.Messages.importMessagesDirectory(__dirname);
// Load the specific messages for this file. Messages from @salesforce/command, @salesforce/core,
// or any library that is using the messages framework can also be loaded this way.
// const messages = Messages.loadMessages('sfpm', 'org');
const staticresourceTemplate = `<?xml version="1.0" encoding="UTF-8"?>
<StaticResource xmlns="http://soap.sforce.com/2006/04/metadata">
<cacheControl>Public</cacheControl>
<contentType>application/zip</contentType>
</StaticResource>`;
let _packageJson;
// @ts-ignore
let _dependencyPackageJson;
let _debug;
let _verbose;
function verbose(message) {
    if (_verbose) {
        console.log(message);
    }
}
function debug(message) {
    if (_debug) {
        console.log(message);
    }
}
function webpackConfigTemplate(resources, flags) {
    const namespace = flags.namespace;
    return `
	//var webpack = require("webpack");

	module.exports = {
		mode: ${flags.debug ? '\'development\'' : '\'production\''},
		entry: {
			${resources.map(sr => `'${sr}': './temp_sfpm/${sr}-input.js'`)
        .join(',')}
		},
		module: {
			rules: [
				{
					${false ? 'parser: { amd: false },' : ''}
					test: /\.js$/,
					exclude: /(node_modules|bower_components)/
				}
			]
		},
		${false ? `plugins: [
			new webpack.ProvidePlugin({
				'lodash': 'lodash'
			})
		],` : ''}
		output: {
			filename: '[name].js',
			${namespace && false ? `library: '${namespace}',` : ''}
			${namespace && false ? `libraryTarget: 'umd',` : ''}
			path: __dirname + '/temp_sfpm/dist'
		}
	}`;
}
const dirMap = {
    aura: 'AuraDefinitionBundle',
    staticresources: 'StaticResource',
    labels: 'CustomLabel',
    pages: 'ApexPage',
    classes: 'ApexClass',
    lwc: 'LightningComponentBundle'
};
function addToPackageXml(destPath, entry, type) {
    if (entry.endsWith('-meta.xml'))
        return;
    const packxml = convert.xml2js(fs.readFileSync(path.normalize(`${destPath}/package.xml`)), { compact: true });
    //does section exist?
    let section = _.find(packxml.Package.types, { name: { _text: dirMap[type] } });
    if (!section) {
        section = {
            members: [],
            name: { _text: dirMap[type] }
        };
        packxml.Package.types.push(section);
        verbose(`adding section ${dirMap[type]}; json = ${JSON.stringify(section)}`);
    }
    //does entry exist?
    if (!_.find(section.members, { _text: entry }) && section.members._text !== entry) {
        verbose(`Adding ${entry} to section ${type}`);
        if (!section.members.push) {
            section.members = [section.members];
        }
        section.members.push({ _text: entry });
        fs.writeFileSync(path.normalize(`${destPath}/package.xml`), convert.js2xml(packxml, { compact: true, spaces: 4 }));
    }
    else {
        verbose(`Found ${entry} in section ${type}; no-op`);
    }
}
function mergeLabels(destPath, srcLabelsDir) {
    const destLabelsDir = path.normalize(`${destPath}/labels`);
    if (fs.existsSync(srcLabelsDir)) {
        debug('merging labels');
        let destLabelFilename = path.normalize(`${destLabelsDir}/` + fs.readdirSync(destLabelsDir)[0]);
        const destLabels = convert.xml2js(fs.readFileSync(destLabelFilename), { compact: true });
        forEachFileInDir(srcLabelsDir, labelFile => {
            const srcLabels = convert.xml2js(fs.readFileSync(path.normalize(`${srcLabelsDir}/${labelFile}`)), { compact: true });
            //copy into target label file
            if (!(srcLabels.CustomLabels.labels instanceof Array)) {
                srcLabels.CustomLabels.labels = [srcLabels.CustomLabels.labels];
            }
            srcLabels.CustomLabels.labels.forEach(srcLabel => {
                if (!_.find(destLabels.CustomLabels.labels, { fullName: srcLabel.fullName })) {
                    destLabels.CustomLabels.labels.push(srcLabel);
                    addToPackageXml(destPath, srcLabel.fullName._text, 'labels');
                }
            });
            // srcLabels.elements[0].elements.forEach(srcLabel => {
            //     destLabels.elements[0].elements.push(srcLabel);
            //     addToPackageXml(destPath, srcLabel, 'labels');
            // });
        });
        fs.writeFileSync(destLabelFilename, convert.js2xml(destLabels, { compact: true, spaces: 4 }));
    }
    else {
        verbose(`${srcLabelsDir} doesn't exist`);
    }
}
function forEachFileInDir(path, callback) {
    if (fs.existsSync(path)) {
        fs.readdirSync(path)
            .forEach(file => callback(file));
    }
}
function copyIntoTreeSfdx(srcPath, destPath, packagingVer, fromPackagingVer) {
    verbose('deploying');
    if (packagingVer === 1) {
        if (fromPackagingVer === 2) {
            verbose(`deploying sfdx package into non-sdfx structure; destPath = ${destPath}; srcPath = ${srcPath}`);
            _.each(fs.readdirSync(srcPath), subdir => {
                _.each(fs.readdirSync(path.normalize(`${srcPath}/${subdir}`)), ssdir => {
                    //first Aura
                    forEachFileInDir(path.normalize(`${srcPath}/${subdir}/${ssdir}/aura`), auraBundle => {
                        copyIntoTree(path.normalize(`${srcPath}/${subdir}/${ssdir}/aura/${auraBundle}`), path.normalize(`${destPath}/aura/${auraBundle}`));
                        addToPackageXml(destPath, auraBundle, 'aura');
                    });
                    forEachFileInDir(path.normalize(`${srcPath}/${subdir}/${ssdir}/lwc`), cmp => {
                        copyIntoTree(path.normalize(`${srcPath}/${subdir}/${ssdir}/lwc/${cmp}`), path.normalize(`${destPath}/lwc/${cmp}`));
                        addToPackageXml(destPath, cmp, 'lwc');
                    });
                    //Next Classes (and probably treat the rest the same too)
                    forEachFileInDir(path.normalize(`${srcPath}/${subdir}/${ssdir}/classes`), clazz => {
                        copyIntoTree(path.normalize(`${srcPath}/${subdir}/${ssdir}/classes/${clazz}`), path.normalize(`${destPath}/classes/${clazz}`));
                        addToPackageXml(destPath, clazz.replace('.cls', ''), 'classes');
                    });
                    forEachFileInDir(path.normalize(`${srcPath}/${subdir}/${ssdir}/pages`), page => {
                        copyIntoTree(path.normalize(`${srcPath}/${subdir}/${ssdir}/pages/${page}`), path.normalize(`${destPath}/pages/${page}`));
                        addToPackageXml(destPath, page.replace('.page', ''), 'pages');
                    });
                    forEachFileInDir(path.normalize(`${srcPath}/${subdir}/${ssdir}/staticresources`), sr => {
                        //zip up the expanded folder into the pack 1.0 structure
                        if (!sr.includes('-meta.xml')) {
                            const output = fs.createWriteStream(path.normalize(`${destPath}/staticresources/${sr}.resource`));
                            const archive = archiver('zip');
                            archive.pipe(output);
                            const pathToZip = path.normalize(`${srcPath}/${subdir}/${ssdir}/staticresources/${sr}`);
                            archive.directory(pathToZip, '');
                            archive.finalize();
                            addToPackageXml(destPath, sr, 'staticresources');
                        }
                    });
                    mergeLabels(destPath, path.normalize(`${srcPath}/${subdir}/${ssdir}/labels`));
                });
            });
        }
        else {
            //packager has a 1.0 project
            verbose(`Deploying non-sfdx package into non-sfdx structure`);
            forEachFileInDir(path.normalize(`${srcPath}/aura`), auraBundle => {
                copyIntoTree(path.normalize(`${srcPath}/aura/${auraBundle}`), path.normalize(`${destPath}/aura/${auraBundle}`));
                addToPackageXml(destPath, auraBundle, 'aura');
            });
            forEachFileInDir(path.normalize(`${srcPath}/lwc`), component => {
                copyIntoTree(path.normalize(`${srcPath}/lwc/${component}`), path.normalize(`${destPath}/lwc/${component}`));
                addToPackageXml(destPath, component, 'lwc');
            });
            forEachFileInDir(path.normalize(`${srcPath}/classes`), clazz => {
                copyIntoTree(path.normalize(`${srcPath}/classes/${clazz}`), path.normalize(`${destPath}/classes/${clazz}`));
                addToPackageXml(destPath, clazz.replace('.cls', ''), 'classes');
            });
            forEachFileInDir(path.normalize(`${srcPath}/pages`), page => {
                copyIntoTree(path.normalize(`${srcPath}/pages/${page}`), path.normalize(`${destPath}/pages/${page}`));
                addToPackageXml(destPath, page.replace('.page', ''), 'pages');
            });
            forEachFileInDir(path.normalize(`${srcPath}/staticresources`), sr => {
                copyIntoTree(path.normalize(`${srcPath}/staticresources/${sr}`), path.normalize(`${destPath}/staticresources/${sr}`));
                addToPackageXml(destPath, sr.replace('.resource', ''), 'staticresources');
            });
            mergeLabels(destPath, path.normalize(`${srcPath}/labels`));
        }
    }
    else { // target is sfdx
        if (fromPackagingVer === 2) {
            //rename "main" or whatever the top-level is to "sfpm"
            _.each(fs.readdirSync(srcPath), subdir => copyIntoTree(path.normalize(`${srcPath}/${subdir}`), path.normalize(`${destPath}/sfpm_modules`)));
        }
        else {
            _.each(fs.readdirSync(srcPath), subdir => copyIntoTree(path.normalize(`${srcPath}/${subdir}`), path.normalize(`${destPath}/src`)));
            // todo expand static resource zip
            // const extract = require('extract-zip')
            // extract(source, {dir: target}, function (err) {
            //     // extraction is complete. make sure to handle the err
            // })
        }
    }
}
function copyIntoTree(srcPath, destPath) {
    fs.copySync(srcPath, destPath);
}
function getAllJSDeps(packageJson) {
    return _.chain(_getAllJSDeps(packageJson, true))
        .flattenDeep()
        // .filter(dep => dep !== '')
        .value();
}
function _getAllJSDeps(packageJson, rootModule) {
    return (packageJson.staticresource || []).concat(
    //if you have deps & you're an sfpm module or the root module, recurse
    packageJson.dependencies && (packageJson.sfpm || rootModule)
        ? _.map(Object.keys(packageJson.dependencies), dep => {
            const depJson = JSON.parse(fs.readFileSync(path.normalize(`./node_modules/${dep}/package.json`), 'utf8'));
            return _getAllJSDeps(depJson, false);
        })
        : []);
}
function copySfdcCode() {
    debug('copy sfdc code into tree for ' + _packageJson.name);
    if (fs.existsSync(path.normalize('./sfdx-project.json'))) {
        verbose('\ttarget is sfdx');
        const sfdxJson = JSON.parse(fs.readFileSync(path.normalize(`./sfdx-project.json`), 'utf8'));
        verbose('\tparsed JSON');
        const proj = _.find(sfdxJson.packageDirectories, { default: true });
        verbose('\tFound the default project');
        _copySfdcCode(proj.path, 2);
    }
    else {
        //it's a packaging 1.0 codebase
        verbose('\ttarget is non-sfdx');
        _copySfdcCode(path.normalize('./src'), 1);
    }
}
function _copySfdcCode(targetPath, packagingVer) {
    packagingVer = packagingVer || 2;
    verbose(`dependencies = ${_packageJson.dependencies}`);
    _.each(Object.keys(_packageJson.dependencies || []), dep => {
        const depJson = JSON.parse(fs.readFileSync(path.normalize(`./node_modules/${dep}/package.json`), 'utf8'));
        if (depJson.sfpm) {
            //todo iterate over all modules inside sfdx project
            console.log(`sfpm: ${depJson.name} v${depJson.version}`);
            _dependencyPackageJson = depJson;
            let projectPath;
            let fromPackagingVer = 2;
            const packProjJson = path.normalize(`./node_modules/${dep}/sfdx-project.json`);
            verbose(`packProjJson = ${packProjJson}`);
            if (fs.existsSync(packProjJson)) {
                verbose(`\t${dep} is sfdx`);
                const sfdxJson = JSON.parse(fs.readFileSync(packProjJson, 'utf8'));
                const proj = _.find(sfdxJson.packageDirectories, { default: true });
                projectPath = proj.path;
            }
            else {
                verbose(`\t${dep} is non-sfdx`);
                projectPath = 'src';
                fromPackagingVer = 1;
            }
            copyIntoTreeSfdx(path.normalize(`./node_modules/${depJson.name}/${projectPath}`), path.normalize(`./${targetPath}`), packagingVer, fromPackagingVer);
        }
    });
}
function isSfpm(moduleName) {
    return !!(JSON.parse(fs.readFileSync(path.normalize(`./node_modules/${moduleName}/package.json`), 'utf8')).sfpm);
}
function doBuild(context) {
    _debug = context.flags.debug || context.flags.verbose;
    _verbose = context.flags.verbose;
    if (!context.flags.skip_npm) {
        const cmd = 'npm install -s ' + (context.args.package || '')
            + (context.flags.save ? ' --save' : '');
        debug(`executing: ${cmd}`);
        child_process.execSync(cmd, { stdio: [0, 1, 2] });
        debug("--------------\nnpm install complete");
        console.log('\n----------------------------------------');
    }
    //create a scratch dir
    const tempDir = path.normalize('./temp_sfpm');
    fs.removeSync(tempDir);
    fs.mkdirSync(tempDir);
    const packJson = JSON.parse(fs.readFileSync(path.normalize('./package.json'), 'utf8'));
    packJson.staticresource = packJson.staticresource || [];
    //add new dep to .staticresource
    if (context.flags.save && context.args.package && !isSfpm(context.args.package)) {
        if (!packJson.staticresource.includes(context.args.package)) {
            packJson.staticresource.push(context.args.package);
            fs.writeFileSync(path.normalize(`./package.json`), JSON.stringify(packJson, null, 4));
        }
    }
    //if dependency isn't sfpm, assume it's javascript and needs to be a staticresource
    if (!packJson.staticresource.includes(context.args.package) && context.args.package && !isSfpm(context.args.package)) {
        packJson.staticresource.push(context.args.package);
    }
    //copy SFDC aura/lwc/class etc files to /sfpm
    _packageJson = packJson;
    copySfdcCode();
    const dependentJSmodules = getAllJSDeps(packJson);
    if (dependentJSmodules.length > 0) {
        debug(`generating webpack entry-points for: ${dependentJSmodules}`);
        let webpackContents;
        const jsNamespace = context.flags.namespace ? context.flags.namespace + '_' : '';
        dependentJSmodules.forEach(sr => {
            const globalName = `${jsNamespace}${sr.replace(/[-]/g, '')}`;
            webpackContents = `window.${globalName} = require('${sr}')`;
            fs.writeFileSync(path.normalize(`${tempDir}/${sr}-input.js`), webpackContents);
        });
        // const cmd = 'npm install webpack'; ///???
        // console.log(`executing: ${cmd}`);
        // child_process.execSync(cmd,{stdio:[0,1,2]});
        //generate webpack.config.js
        webpackContents = webpackConfigTemplate(dependentJSmodules, context.flags);
        fs.writeFileSync(path.normalize(`./webpack.config.js`), webpackContents);
        const cmd = 'webpack';
        debug(`executing: ${cmd}`);
        child_process.execSync(cmd, { stdio: [0, 1, 2] });
        if (!context.flags.debug) {
            fs.unlinkSync(path.normalize('./webpack.config.js'));
        }
        const sfdxProject = JSON.parse(fs.readFileSync('sfdx-project.json', 'utf8'));
        const proj = _.find(sfdxProject.packageDirectories, { default: true });
        //zip up all the output from webpack into 1 static resource
        fs.ensureDirSync(proj.path + path.normalize('/sfpm_modules/default/staticresources'));
        const output = fs.createWriteStream(path.normalize(`${proj.path}/sfpm_modules/default/staticresources/SFPM.resource`));
        const archive = archiver('zip');
        // listen for all archive data to be written
        // 'close' event is fired only when a file descriptor is involved
        output.on('close', function () {
            if (!context.flags.debug) {
                fs.removeSync(tempDir);
            }
        });
        archive.pipe(output);
        archive.directory(path.normalize('./temp_sfpm/dist'), '');
        archive.finalize();
        fs.writeFileSync(path.normalize(`${proj.path}/sfpm_modules/default/staticresources/SFPM.resource-meta.xml`), staticresourceTemplate);
    }
    else {
        //cleanup
        if (!context.flags.debug) {
            fs.removeSync(tempDir);
        }
    }
}
exports.doBuild = doBuild;
//# sourceMappingURL=sfpm.js.map